/*

--_-_-_-_---
   -_-_-_
    -_-_-
     -__-
    _-_
   _-
   -_
    _-_
    
CyClone Weather Station (Relay Client)

Author: NLD

Date: 2023/Apr/01

Version: 0.4

This code is intended to be for the barebone Atmega 328P or the Arduino Promicro.

This is a stripped down version of the Client.

The only function of this receiver is to turn a relay on or off depending on the outside sunlight.

It ignores any other value but the LightAVG coming from the base.

Regardless that this could use the only hardware UART on the Atmega 328P, I prefer keeping that for debugging.


v0.4: 
 - Public release

v0.3: 
 - RX indicator, error indicator leds removed
 - The received light_avg number is now also validated (fail counter criteria 3)
 - light_avg increased to 400 as it plays better with the current setup 

v0.2:
 - Motion code added from miniphoenix 4

v0.1:
- Forked from cyclone_weather_rx

..............................................................*/
//#define DEBUG

#define weather_rx_relay_ver 0.3
#define LIGHT_THRESHOLD 400

/* HC12 Radio */
#include <SoftwareSerial.h>
SoftwareSerial HC12(9, 10); // HC-12 TX Pin, HC-12 RX Pin                                                                                                                                                                                

/* Motion PIR related */
uint32_t hold_time=300000;  // 300000 Light turn on time on motion -> 5 minutes: Put real value here not calculation
                   
//Pins
#define motionPin1 5
#define motionPin2 6
#define bypass A3
int on_led=7;
int off_led=8;
int dark_led=4;     
int relayPin=A2;

//Timing
bool on_hold=false;
uint32_t curr_time;
uint32_t last_event;
unsigned int l;

//Commands
char Command_PC[64], Command_HC12[64], Command[64];
static int Count_PC = 0;
static int Count_HC12 = 0;
static bool Complete_PC = false;
static bool Complete_HC12 = false;

unsigned long fc = 0;

int ON=HIGH;
int OFF=LOW;
int LON=LOW;
int LOFF=HIGH;

// Weather telemetry
float temp_in  = 0.00;
float hum_in   = 0.00;

float temp_out = 0.00; 
float hum_out  = 0.00;
float light    = 0.00;
float light_avg= LIGHT_THRESHOLD+1; // We don't want it to turn on at start
float rain     = 0.00;
float wind     = 0.00;
float pressure = 0.00;

unsigned int windiri=0; // 0 - North

char Msg[150];
char MsgReceived[150]; 


// Searches a string for a pattern character
// returns 0 - if pattern not found
// returns x - the number of occasions the pattern is found
int StrContains(char *str, char *pattern)
{
    char index = 0;
    char len;
    unsigned int occurrence = 0;
    
    len = strlen(str);
    
    if (strlen(pattern) > len) {
        return 0;
    }
    while (index < len) {
        if (str[index] == pattern[0])
            occurrence++;

        index++;
    }

     return occurrence;
}

void str_replace(char *src, char *oldchars, char *newchars) { // utility string function
  char *p = strstr(src, oldchars);
  char buf[100];
  do {
    if (p) {
      memset(buf, '\0', strlen(buf));
      if (src == p) {
        strcpy(buf, newchars);
        strcat(buf, p + strlen(oldchars));
      } else {
        strncpy(buf, src, strlen(src) - strlen(p));
        strcat(buf, newchars);
        strcat(buf, p + strlen(oldchars));
      }
      memset(src, '\0', strlen(src));
      strcpy(src, buf);
    }
  } while (p && (p = strstr(src, oldchars)));
}

/* Incoming command handling */
void xcomm() {

#ifdef DEBUG
/* COMM FROM PC */   
if (Serial.available()){

    Serial.println("PC: New command, collecting...");
    memset(Command_PC, 0, sizeof(Command_PC));
    while (Serial.available()){
      char character = Serial.read();
      if (character=='\n')
         {
          Complete_PC =true;
          Count_PC = 0;
         } else {
             if (Count_PC < 62) { // Don't overflow the array...!
                 Command_PC[Count_PC++] = character;
                 Command_PC[Count_PC] = 0;
             }
         }
  }
}   
#endif

/* COMM FROM HC12 */   
if (HC12.available()){

  #ifdef DEBUG
    Serial.print("HC12: ");
  #endif
    memset(Command_HC12, 0, sizeof(Command_HC12));
    while (HC12.available()){
      char character = HC12.read();
      delay(100); /* DO NOT REMOVE THIS WITH LOW BPS like 1200Bps or expect large amount of broken data */
      if (character=='\n')
         {
          Complete_HC12 =true;          
          Count_HC12 = 0;
         } else {
             if (Count_HC12 < 62) { // Don't overflow the array...!
                 Command_HC12[Count_HC12++] = character;
                 Command_HC12[Count_HC12] = 0;
             }
         }
  }
}   
  

/* Command_PC is just for debugging, disabled normally */
 if (Complete_PC||Complete_HC12) {
  
  strcpy(Command, Command_HC12);
  
  #ifdef DEBUG
  if (Complete_HC12)
  {
    strcpy(Command, Command_HC12);
    Serial.print("Command received HC12: |"); // For debug we use the | as xmit begin/stop indicator as the remote will send > <
    Serial.print(Command);
    Serial.println("|");
  }
  
  if (Complete_PC)
  {
    strcpy(Command, Command_PC);
    Serial.print("Command received PC (overwriting HC12): |");
    Serial.print(Command);
    Serial.println("|");
  }
  #endif  

/*  
 * 8 Data points: >temperature,humidity,lightsensitivity,light_avg,pressure,wind,winddirection,rain<
 * From these all is float except the windirection which is an unsigned int between 0-15

   Step 1: look for both of the >markers< neededs to be present 
   Step 2: 8 comma separated values needs to be present
*/
  if (StrContains(Command,">") == 1 && StrContains(Command,"<") == 1 && StrContains(Command,",") == 7) {
    
                    // Step 3: remove markers
                    str_replace(Command, ">", "");
                    str_replace(Command, "<", "");


                    
                    temp_out = atof(strtok(Command, ","));
                    hum_out  = atof(strtok(NULL, ","));
                    light    = atof(strtok(NULL, ","));
                    light_avg= atof(strtok(NULL, ","));
                    pressure = atof(strtok(NULL, ","));                   
                    wind     = atof(strtok(NULL, ","));
                    windiri  = atoi(strtok(NULL, ","));
                    rain     = atof(strtok(NULL, ","));
                    
                  #ifdef DEBUG
                    Serial.println("Data decoded: ");
                    Serial.print("var1 temp: >");
                    Serial.println(temp_out);                  
                    Serial.print("var2 hum: >");
                    Serial.println(hum_out);                  
                    Serial.print("var3 light: >");
                    Serial.println(light);                  
                    Serial.print("var4 light_avg: >");
                    Serial.println(light_avg);                        
                    Serial.print("var5 pressure: >");
                    Serial.println(pressure);                  
                    Serial.print("var6 wind: >");
                    Serial.println(wind);                  
                    Serial.print("var7 wind direction: >");
                    Serial.println(windiri);                  
                    Serial.print("var8 rain amount: >");
                    Serial.println(rain);                  
                  #endif
                  //fc=0; // Comment this out for regression testing and leave it for days
   } else if (strcmp(Command, "LIGHT_ON") == 0) {
            on();
   }       
     else if (strcmp(Command, "LIGHT_OFF") == 0) {
            off();
   } else if (strlen(Command) < 5) {
     fc++;
     // Some absolute junk completely ignore this don't even alert
   } else { // The base station don't accept incoming commands, in case of broken xmit no need to cry about it.   
       fc++;
       #ifdef DEBUG
          Serial.println("CYCLONE_CLIENT_ERR_UNK_DATA");            
       #endif    
   }
                        
                       
       Complete_PC=false;
       Complete_HC12=false;
 }
 
}

void on()
{
 digitalWrite(relayPin, ON);
 digitalWrite(on_led, ON);
 digitalWrite(off_led, OFF);

  #ifdef DEBUG  
   Serial.println("CYCLONE_LIGHT_ON");
  #endif
}

void off()
{
 digitalWrite(relayPin, OFF);
 digitalWrite(on_led, OFF);
 digitalWrite(off_led, ON);

  #ifdef DEBUG  
   Serial.println("CYCLONE_LIGHT_OFF");
  #endif
}

void setup() {
 pinMode(relayPin,OUTPUT);
 pinMode(on_led, OUTPUT);       // Digital
 pinMode(off_led, OUTPUT);      // Digital 
 pinMode(dark_led, OUTPUT);     // Digital 
 digitalWrite(on_led, OFF);
 digitalWrite(off_led, OFF);
 digitalWrite(dark_led, OFF);

 pinMode(motionPin1, INPUT);    // Digital
 pinMode(motionPin2, INPUT);    // Digital
 pinMode(bypass, INPUT);        // Digital  


#ifdef DEBUG 
  Serial.begin(9600);  // PC comm
#endif  
  HC12.begin(1200);    // HC12 radio comm 

// System bootup indicator
for (unsigned int i=0;i<5;i++) {
   digitalWrite(on_led, ON);
   digitalWrite(off_led, ON);
   digitalWrite(dark_led, ON);
   delay(1000);
   digitalWrite(on_led, OFF);
   digitalWrite(off_led, OFF);
   digitalWrite(dark_led, OFF);
   delay(1000);   
}

#ifdef DEBUG
  Serial.print("Cyclone Relay Client starting up ... [ Ver: ");
  Serial.print(weather_rx_relay_ver);
  Serial.println(" ]");
#endif

 off();
}  

void loop() {

// Bypass switch basically makes the receiver ignore the light data from the station and will act on the motion anyway.
if (digitalRead(bypass) == LOW)
  light_avg=99; // Fake it (ON)
else
{
  xcomm(); // Collect data on radio
  
  //Third validation on the received data
  if (light_avg < 0 || light_avg > 100000) {
   fc++;
   light_avg=LIGHT_THRESHOLD+1; // Fake it (OFF)
  }
}


curr_time = millis();

// Limit light power on events
if (on_hold && curr_time - last_event >= hold_time) {        
     on_hold = false;
}

  

if (light_avg<LIGHT_THRESHOLD) 
 digitalWrite(dark_led, ON);
else
 digitalWrite(dark_led, OFF);
 
//Motion detected   
   if ( digitalRead(motionPin1) == HIGH || digitalRead(motionPin2) == HIGH )
        {          
            if (!on_hold && light_avg<LIGHT_THRESHOLD)
            {              
               on();               
               #ifdef DEBUG              
                Serial.println("CYCLONE_MOTION_DETECTED_NIGHT");            
               #endif
            }
            
            last_event = curr_time; // Bumps up the hold timer
            on_hold = true;

        }

//Even if the hold timer expired if there was motion just before this will lock again     
  if (!on_hold)
  {          
          #ifdef DEBUG
            Serial.println("CYCLONE_TURNING_OFF");            
          #endif
          
          off();
     delay(1000);   
  }
  
          #ifdef DEBUG
            Serial.print("LOOP:");            
            Serial.println(l++);
            Serial.print("TIMER: ");
            Serial.println(curr_time - last_event);
            Serial.print("Sensor 1 reading: ");
            Serial.println(digitalRead(motionPin1));
            Serial.print("Sensor 2 reading: ");
            Serial.println(digitalRead(motionPin2));   
            Serial.print("Last AVG Light data: ");
            Serial.println(light_avg);
            Serial.print("Failure Count: ");
            Serial.println(fc);            
            Serial.print("Current state: ");
            if (on_hold) 
              Serial.println(" [LOCKED]");         
            else 
              Serial.println(" [CLEAR]");       
            
            Serial.print("Bypass: ");
            
            if(digitalRead(bypass) == LOW)
              Serial.println(" [ACTIVE]");
            else
              Serial.println(" [INACTIVE]");      
          #endif

delay(1000);

}
